package ch.hslu.swk.g03.logger.server;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;

import java.net.Socket;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;

import ch.hslu.loggerinterface.Loglevel;
import ch.hslu.swk.g03.logger.common.ConverterUtils;
import ch.hslu.swk.g03.logger.common.LogMessage;

/**
 * ServerTCP Client Thread.
 *
 * @author Felix Rohrer <felix.rohrer@stud.hslu.ch>
 */
public class ServerTCPConnThread extends Thread {

    private static final int READ_CYCLE = 100;
    private static final int MAX_CYCLES = 10;
    private static final int SEMAPHORE_TIMEOUT = 2; // seconds
    private static final int THREAD_TIMEOUT = 1; // seconds
    private static final int MSG_HEADER_SIZE = 4; // byte

    private final Socket socket;
    private final ServerTCP server;
    private boolean listening = true;
    private static Semaphore syncSemaphore = new Semaphore(1);

    /**
     * Default Constructor.
     *
     * @param server
     *            TCP-Server (Parent-Thread)
     * @param socket
     *            Client Socket
     */
    public ServerTCPConnThread(final ServerTCP server, final Socket socket) {
	super("ServerTCPConnThread");
	this.server = server;
	this.socket = socket;
    }

    /**
     * ServerTCP Connection Thread Start Method.
     */
    public final void run() {
	// let run only one thread at once --> sequential logging!
	try {
	    // try acquire the semaphore within SEMAPHORE_TIMEOUT seconds - otherwhise terminate this thread
	    if (!syncSemaphore.tryAcquire(SEMAPHORE_TIMEOUT, TimeUnit.SECONDS)) {
		return;
	    }
	} catch (InterruptedException e) {
	    System.err.println("[ERROR:ServerTCPConnThread] Semaphore TimeOut, Client coulnd't connect: "
		    + socket.getRemoteSocketAddress());
	    return;
	}

	InputStream inStream = null;

	server.addListeningClient(this);
	System.out.println("[INFO:ServerTCPConnThread] new Client connected: " + socket.getRemoteSocketAddress());
	super.setName(socket.getRemoteSocketAddress().toString());

	try {
	    inStream = socket.getInputStream();
	    final ByteArrayOutputStream byteStream = new ByteArrayOutputStream();

	    int cycles = 0;
	    byte[] buffer = null;
	    int size = 0;
	    int length = 0;
	    while (listening) {
		if (inStream.available() >= MSG_HEADER_SIZE) {
		    try {
			// read the first 4 (MSG_HEADER_SIZE) byte -> Message length
			buffer = new byte[MSG_HEADER_SIZE];
			inStream.read(buffer);

			// Get the size of the log message
			size = ConverterUtils.byteArrayToInt(buffer);

			// abort if size < 0
			if (size < 0) {
			    break;
			}

			// reset byteStream an read remaining bytes --> LogMessage
			byteStream.reset();
			buffer = new byte[size];
			length = inStream.read(buffer);
			// copy buffer to byteStream for deserialize
			byteStream.write(buffer, 0, length);

			// Deserialize LogMessage from ByteStream
			final LogMessage message = (LogMessage) ConverterUtils.deserialize(byteStream.toByteArray());

			// Do not log the connection Test
			if (message.getLoglevel() == Loglevel.OFF) {
			    System.out.println("[INFO:ServerTCPConnThread] Connection Test...");
			} else {
			    LoggerDB.getInstance().addMessage(message);
			}

		    } catch (ClassNotFoundException e) {
			System.err.println("[ERROR:ServerTCPConnThread] deserialize -> ClassNotFoundException...");
			// force disconnect client
			listening = false;
		    }
		}
		// abort if no LogMessage is received within timeout
		if (cycles++ >= MAX_CYCLES) {
		    System.err.println("[ERROR:ServerTCPConnThread] Kill this thread, "
			               + "client hasn't sent any LogMessage within "
			               + THREAD_TIMEOUT + " second!");
		    break;
		}

		try {
		    Thread.sleep(READ_CYCLE);
		} catch (InterruptedException e) {
		    System.err.println("[ERROR:ServerTCPConnThread] Sleep InterruptedException...");
		}
	    }
	    byteStream.close();

	} catch (IOException e) {
	    e.printStackTrace();
	} finally {
	    // close stream, socket
	    try {
		if (inStream != null) {
		    inStream.close();
		}
		socket.close();
	    } catch (IOException e) {
		System.err.println("[ERROR:ServerTCPConnThread] couldn't close inStream / socket...");
	    } finally {
		System.out.println("[INFO:ServerTCPConnThread] Client disconnect: " + socket.getRemoteSocketAddress());
		server.removeListeningClient(this);
	    }

	    // Release Semaphore
	    syncSemaphore.release();
	}
    }
}
